//
// Author:    Jae Chung
// Data:      10/10/99
// File Name: mm-app-mpeg.cc
// 

#define INIT_SCALE 0
#define MAX_SCALE  4

#include "random.h"
#include "mm-app-mpeg.h"
#include <string.h>


// MmAppMpeg OTcl linkage class
static class MmAppMpegClass : public TclClass {
 public:
  MmAppMpegClass() : TclClass("Application/MmAppMpeg") {}
  TclObject* create(int, const char*const*) {
    return (new MmAppMpeg);
  }
} class_app_mpeg;


// When snd_timer_ expires call MmAppMpeg:send_mm_pkt()
void MmAppMpegSendTimer::expire(Event*)
{
  t_->send_mm_pkt();
}


// When ack_timer_ expires call MmAppMpeg:send_ack_pkt()
void MmAppMpegAckTimer::expire(Event*)
{
  t_->send_ack_pkt();
}


// Constructor (also initialize instances of timers)
MmAppMpeg::MmAppMpeg() : running_(0), snd_timer_(this), ack_timer_(this)
{
  bind_bool("flow_control_", &flow_control_);
  bind("frame_per_sec_", &frame_per_sec_);
  bind("max_interval_", &max_interval_);

  fdr_ = NULL;
  fdw_ = NULL;
  file_closed_ = 0;
  init_recv_pkt_accounting();
}


// OTcl command interpreter
int MmAppMpeg::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();

  if (argc == 3) {
    // Attach Agent
    if (strcmp(argv[1], "attach-agent") == 0) {
      agent_ = (Agent*) TclObject::lookup(argv[2]);
      if (agent_ == 0) {
	tcl.resultf("no such agent %s", argv[2]);
	return(TCL_ERROR);
      }

      // Make sure the underlying agent support MM
      if(agent_->supportMM()) {
	agent_->enableMM();
	sup_mm_ = 1;
      }
      else {
	sup_mm_ = 0;
	// tcl.resultf("agent \"%s\" does not support MM Application", argv[2]);
	// return(TCL_ERROR);
      }
      
      agent_->attachApp(this);
      return(TCL_OK);
    }
    // Get Mpeg Trace Input File Name
    if (strcmp(argv[1], "mpeg-trace-input") == 0) {
      if ((fdr_ = fopen(argv[2], "r")) == NULL) {
	tcl.resultf("cannot open mpeg-trace-input file \"%s\"", argv[2]);
        return(TCL_ERROR);
      }
      return(TCL_OK);
    }
    // Record MM Packet Arival to A File
    if(strcmp(argv[1], "record-mm-packet-arrival") == 0) {
      if((fdw_ = fopen(argv[2], "w")) == NULL) {
	tcl.resultf("cannot create mm-packet-arrival file \"%s\"", argv[2]);
	return(TCL_ERROR);
      }
      return(TCL_OK);
    }
  }
  return (Application::command(argc, argv));
}



void MmAppMpeg::init()
{
  seq_ = 0;             // MM sequence number (start from 0)
  fnum_ = -1;            // MM frame number (start from 0)

  if(flow_control_)
    scale_ = INIT_SCALE;  // Start at given scale 
  else
    scale_ = MAX_SCALE;   // Start with Max Scale

  p_pkt_sent_ = 0;      // Flag for IBBPBBI format: drop every 
                        // other 'P' packet at scale 1.
  interval_ = (double)(1/(double)frame_per_sec_);

  recv_pkt_type[0] = 'B';
  recv_pkt_type[1] = 'B';
  recv_pkt_type[2] = 'P';
  recv_pkt_type[3] = 'B';
  recv_pkt_type[4] = 'B';
  recv_pkt_type[5] = 'P';
  recv_pkt_type[6] = 'B';
  recv_pkt_type[7] = 'B';
  recv_pkt_type[8] = 'I';
}


void MmAppMpeg::start()
{
  if(fdr_ == NULL) {
    printf ("MmAppMpeg Error: specify mpeg-trace-input file.\n");
    exit(1);
  }
  init();
  running_ = 1;
  send_mm_pkt();
}


void MmAppMpeg::stop()
{
  running_ = 0;
  if (file_closed_ == 0) {
    fclose(fdr_);
    file_closed_ = 1;
  }
}


// Send application data packet
void MmAppMpeg::send_mm_pkt()
{
  hdr_mm mh_buf;
  char ftype;

  if (running_) {
    // Get Size of next MPEG Packet
    if ((pktsize_ = get_pkt_size(&ftype)) == -1) {
      stop();       // if EOF then stop
      return;
    }

    // Increment frame number 
    // (account for not sent frames also)
    fnum_++;

    // pktsize_ = (positive integer, 0, -1)
    // where 0 means due to network congestion, sender will not
    // send the packet.

    if (pktsize_ > 0) {
      if(sup_mm_) {
	// the below info is passed to UDPmm agent, which will write it 
	// to MM header after packet creation.
	mh_buf.ack = 0;            // This is a MM packet
	mh_buf.seq = seq_++;       // MM sequece number
	mh_buf.nbytes = pktsize_;  // Size of MM packet (NOT UDP packet size)
	mh_buf.ftype = ftype;      // Frame-type (I,B,P)
	mh_buf.fnum = fnum_;
	mh_buf.time = Scheduler::instance().clock(); // Current time
	mh_buf.scale = scale_;                       // Current scale value
	agent_->sendmsg(pktsize_, (char*) &mh_buf);  // send to UDP
      }
      else 
	agent_->sendmsg(pktsize_);
    }

    // Reschedule the send_pkt timer
    snd_timer_.resched(interval_);
  }
}

// Read Next Mpeg packet type and size from input file and
// determine whether or not transmit depending on current scale value
int MmAppMpeg::get_pkt_size(char* ftype)
{
  char pkt_type_;
  int  size_read_;
  int  pkt_size_ = 0;

  // if EOF return -1 to finish
  if (fscanf(fdr_, "%c\t%d\n", &pkt_type_, &size_read_) == EOF) return -1;

  // Different transmission policy for each media scale level
  // Initially, pkt_size_ = 0
  switch (scale_) {
    case 0:
      if (pkt_type_ == 'I') pkt_size_ = size_read_;
      break;

    case 1:
      if (pkt_type_ == 'I') pkt_size_ = size_read_;
      if (pkt_type_ == 'P') {
	if (recv_pkt_type[2] == 'I') {
	  // for IBBPBBPBBI and IBBPBBPBBPBBI format
	  if (recv_pkt_type[8] == 'P') { 
	    pkt_size_ = size_read_;
	  }
	  // for IBBPBBI format
	  else {
	    if (p_pkt_sent_ == 0) {
	      pkt_size_ = size_read_;
	      p_pkt_sent_ = 1;
	    }
	    else
	      p_pkt_sent_ = 0;
	  }
	}
      }
      break;

    case 2:
      if (pkt_type_ == 'I') pkt_size_ = size_read_;
      if (pkt_type_ == 'P') pkt_size_ = size_read_;
      break;

    case 3:
      if (pkt_type_ == 'I') pkt_size_ = size_read_;
      if (pkt_type_ == 'P') pkt_size_ = size_read_;
      if ((pkt_type_ == 'B') && (recv_pkt_type[0] == 'B')) pkt_size_ = size_read_;
      break;

    case 4:
      pkt_size_ = size_read_;
      break;

    default:
      printf("Error: unrecognized mpeg packet type\n");
      exit(1);
      break;
  }

  update_recv_pkt_type(pkt_type_);
  *ftype = pkt_type_;
  return pkt_size_;
}


// Keep track of 9 previously read packet type.
void MmAppMpeg::update_recv_pkt_type(char pkt_type)
{
  int i;
  for (i=8; i>0; i--) {
    recv_pkt_type[i] = recv_pkt_type[i-1];
  }
  recv_pkt_type[0] = pkt_type;
}



// Receive message from underlying agent
void MmAppMpeg::recv_msg(int nbytes, const char *msg = 0)
{
  if(msg) {
    hdr_mm* mh_buf = (hdr_mm*) msg;

    if(mh_buf->ack == 1) {
      // If received packet is ACK` packet and enable flow_control
      if(flow_control_) set_scale(mh_buf);
    }
    else {
      // If received packet is MM packet
      account_recv_pkt(mh_buf);
      if(p_accnt.tot_recv_pkts == 1) ack_timer_.resched(p_accnt.rtt);
      if(p_accnt.lost_pkts > 0) send_ack_pkt();
    }
  }
}


// Sender sets its scale to what reciver notifies
void MmAppMpeg::set_scale(const hdr_mm *mh_buf)
{
  scale_ = mh_buf->scale;
}


void MmAppMpeg::init_recv_pkt_accounting()
{
  p_accnt.last_seq = -1;
  p_accnt.last_scale = INIT_SCALE;
  p_accnt.last_time = 0;
  p_accnt.lost_pkts = 0;
  p_accnt.recv_pkts = 0;
  p_accnt.tot_recv_pkts = 0;
}


void MmAppMpeg::account_recv_pkt(const hdr_mm *mh_buf)
{ 
  double local_time = Scheduler::instance().clock();

  // Count Received packets and Calculate Packet Loss
  p_accnt.tot_recv_pkts ++;
  p_accnt.recv_pkts ++;
  p_accnt.lost_pkts += (mh_buf->seq - p_accnt.last_seq - 1);
  p_accnt.last_seq = mh_buf->seq;
  p_accnt.last_time = local_time;

  // Calculate RTT
  if(p_accnt.tot_recv_pkts == 1)
    p_accnt.rtt = 2*(local_time - mh_buf->time);
  else
    p_accnt.rtt = 0.95 * p_accnt.rtt + 0.05 * 2*(local_time - mh_buf->time); 

  // Record mm-packet-arrival
  if(fdw_ != NULL)
    fprintf(fdw_, "%lf\t%d\t%c\t%lf\n", local_time, mh_buf->fnum, mh_buf->ftype, mh_buf->time);
}


void MmAppMpeg::send_ack_pkt(void)
{
  double local_time = Scheduler::instance().clock();

  adjust_scale();

  // send ack message
  hdr_mm ack_buf;
  ack_buf.ack = 1;  // this packet is ack packet
  ack_buf.time = local_time;
  ack_buf.nbytes = 40;  // Ack packet size is 40 Bytes
  ack_buf.scale = p_accnt.last_scale;
  agent_->sendmsg(ack_buf.nbytes, (char*) &ack_buf);

  // schedul next ACK time
  ack_timer_.resched(p_accnt.rtt);
}


void MmAppMpeg::adjust_scale(void)
{  
  // More than one MM packet arrived
  if(p_accnt.recv_pkts > 0) {
    if(p_accnt.lost_pkts > 0)
      p_accnt.last_scale = (int)(p_accnt.last_scale / 2);
    else {
      p_accnt.last_scale++;
      if(p_accnt.last_scale > 4) p_accnt.last_scale = 4;
    }
  }
  // No MM packet arrived within RTT
  else{
    double local_time = Scheduler::instance().clock();
    if((local_time - p_accnt.last_time) > max_interval_)
      p_accnt.last_scale = (int)(p_accnt.last_scale / 2);
  }
  p_accnt.recv_pkts = 0;
  p_accnt.lost_pkts = 0;
}
